"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---
Define LitPop class.
"""
import logging
from pathlib import Path
import numpy as np
import rasterio
import geopandas
from shapefile import Shape
import shapely
import pandas as pd

import climada.util.coordinates as u_coord
import climada.util.finance as u_fin

from climada.entity.tag import Tag
from climada.entity.exposures.litpop import nightlight as nl_util
from climada.entity.exposures.litpop import gpw_population as pop_util
from climada.entity.exposures.base import Exposures, INDICATOR_IMPF, DEF_REF_YEAR
from climada.util.constants import SYSTEM_DIR
from climada import CONFIG
LOGGER = logging.getLogger(__name__)

GPW_VERSION = CONFIG.exposures.litpop.gpw_population.gpw_version.int()
"""Version of Gridded Population of the World (GPW) input data. Check for updates."""

class LitPop(Exposures):
    """
    Holds geopandas GeoDataFrame with metada and columns (pd.Series) defined in
    Attributes of Exposures class.
    LitPop exposure values are disaggregated proportional to a combination of
    nightlight intensity (NASA) and Gridded Population data (SEDAC).
    Total asset values can be produced capital, population count,
    GDP, or non-financial wealth.

    Calling sequence example:
    country_names = ['CHE', 'Austria']
    exp = LitPop.from_countries(country_names)
    exp.plot()

    Attributes
    ----------
    exponents : tuple of two integers, optional
        Defining powers (m, n) with which lit (nightlights) and pop (gpw)
        go into Lit**m * Pop**n. The default is (1,1).
    fin_mode : str, optional
        Socio-economic value to be used as an asset base that is disaggregated.
        The default is 'pc'.
    gpw_version : int, optional
        Version number of GPW population data, e.g. 11 for v4.11.
        The default is defined in GPW_VERSION.
    """
    _metadata = Exposures._metadata + ['exponents', 'fin_mode', 'gpw_version']

    def set_countries(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_countries instead."""
        LOGGER.warning("The use of LitPop.set_countries is deprecated."
                       "Use LitPop.from_countries instead.")
        self.__dict__ = LitPop.from_countries(*args, **kwargs).__dict__

    @classmethod
    def from_countries(cls, countries, res_arcsec=30, exponents=(1,1),
                       fin_mode='pc', total_values=None, admin1_calc=False,
                       reference_year=DEF_REF_YEAR, gpw_version=GPW_VERSION,
                       data_dir=SYSTEM_DIR):
        """Init new LitPop exposure object for a list of countries (admin 0).
        Sets attributes `ref_year`, `tag`, `crs`, `value`, `geometry`, `meta`,
        `value_unit`, `exponents`,`fin_mode`, `gpw_version`, and `admin1_calc`.

        Parameters
        ----------
        countries : list with str or int
            list containing country identifiers:
            iso3alpha (e.g. 'JPN'), iso3num (e.g. 92) or name (e.g. 'Togo')
        res_arcsec : float, optional
            Horizontal resolution in arc-sec.
            The default is 30 arcsec, this corresponds to roughly 1 km.
        exponents : tuple of two integers, optional
            Defining power with which lit (nightlights) and pop (gpw) go into LitPop. To get
            nightlights^3 without population count: (3, 0).
            To use population count alone: (0, 1).
            Default: (1, 1)
        fin_mode : str, optional
            Socio-economic value to be used as an asset base that is disaggregated
            to the grid points within the country
                * 'pc': produced capital (Source: World Bank), incl. manufactured or
                  built assets such as machinery, equipment, and physical structures
                  `pc` is in constant 2014 USD.
                * 'pop': population count (source: GPW, same as gridded population).
                  The unit is 'people'.
                * 'gdp': gross-domestic product (Source: World Bank) [USD]
                * 'income_group': gdp multiplied by country's income group+1 [USD].
                  Income groups are 1 (low) to 4 (high income).
                * 'nfw': non-financial wealth (Source: Credit Suisse, of households only) [USD]
                * 'tw': total wealth (Source: Credit Suisse, of households only) [USD]
                * 'norm': normalized by country (no unit)
                * 'none': LitPop per pixel is returned unchanged (no unit)
            The default is 'pc'.
        total_values : list containing numerics, same length as countries, optional
            Total values to be disaggregated to grid in each country.
            The default is None. If None, the total number is extracted from other
            sources depending on the value of fin_mode.
        admin1_calc : boolean, optional
            If True, distribute admin1-level GDP (if available). Default: False
        reference_year : int, optional
            Reference year. Default: CONFIG.exposures.def_ref_year.
        gpw_version : int, optional
            Version number of GPW population data.
            The default is GPW_VERSION
        data_dir : Path, optional
            redefines path to input data directory. The default is SYSTEM_DIR.

        Raises
        ------
        ValueError

        Returns
        -------
        exp : LitPop
            LitPop instance with exposure for given countries
        """
        if isinstance(countries, (int, str)):
            countries = [countries] # for backward compatibility

        if total_values is None: # init list with total values per countries
            total_values = [None] * len(countries)
        elif len(total_values) != len(countries):
            raise ValueError("'countries' and 'total_values' must be lists of same length")
        tag = Tag()

        # litpop_list is initiated, a list containing one Exposure instance per
        # country and None for countries that could not be identified:
        if admin1_calc: # each admin 1 region is initiated separately,
                        # with total value share based on subnational GDP share.
                        # This requires GRP (Gross Regional Product) data in the
                        # GSDP data folder.
            litpop_list = [_calc_admin1_one_country(country, res_arcsec, exponents,
                                                    fin_mode, tot_value, reference_year,
                                                    gpw_version, data_dir,
                                                    )
                           for tot_value, country in zip(total_values, countries)]

        else: # else, as default, country is initiated as a whole:
            # loop over countries: litpop is initiated for each individual polygon
            # within each country and combined at the end.
            litpop_list = \
                [cls._from_country(country,
                                   res_arcsec=res_arcsec,
                                   exponents=exponents,
                                   fin_mode=fin_mode,
                                   total_value=total_values[idc],
                                   reference_year=reference_year,
                                   gpw_version=gpw_version,
                                   data_dir=data_dir)
                 for idc, country in enumerate(countries)]
        # make lists of countries with Exposure initaited and those ignored:
        countries_in = \
            [country for lp, country in zip(litpop_list, countries) if lp is not None]
        countries_out = \
            [country for lp, country in zip(litpop_list, countries) if lp is None]
        if not countries_in:
            raise ValueError('No valid country identified in %s, aborting.' % countries)
        litpop_list = [exp for exp in litpop_list if exp is not None]
        if countries_out:
            LOGGER.warning('Some countries could not be identified and are ignored: '
                           '%s. Litpop only initiated for: %s', countries_out, countries_in)

        tag.description = f'LitPop Exposure for {countries_in} at {res_arcsec} as, ' \
                          f'year: {reference_year}, financial mode: {fin_mode}, ' \
                          f'exp: {exponents}, admin1_calc: {admin1_calc}'

        exp = cls(
            data=Exposures.concat(litpop_list).gdf,
            crs=litpop_list[0].crs,
            ref_year=reference_year,
            tag=tag,
            value_unit=get_value_unit(fin_mode),
            exponents = exponents,
            gpw_version = gpw_version,
            fin_mode = fin_mode,
        )

        try:
            rows, cols, ras_trans = u_coord.pts_to_raster_meta(
                (exp.gdf.longitude.min(), exp.gdf.latitude.min(),
                 exp.gdf.longitude.max(), exp.gdf.latitude.max()),
                u_coord.get_resolution(exp.gdf.longitude, exp.gdf.latitude))
            exp.meta = {
                'width': cols,
                'height': rows,
                'crs': exp.crs,
                'transform': ras_trans,
            }
        except ValueError:
            LOGGER.warning('Could not write attribute meta, because exposure'
                           ' has only 1 data point')
            exp.meta = {'crs': exp.crs}
        exp.check()
        return exp

    def set_nightlight_intensity(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_nightlight_intensity instead."""
        LOGGER.warning("The use of LitPop.set_nightlight_intensity is deprecated."
                       "Use LitPop.from_nightlight_intensity instead.")
        self.__dict__ = LitPop.from_nightlight_intensity(*args, **kwargs).__dict__

    @classmethod
    def from_nightlight_intensity(cls, countries=None, shape=None, res_arcsec=15,
                        reference_year=DEF_REF_YEAR, data_dir=SYSTEM_DIR):
        """
        Wrapper around `from_countries` / `from_shape`.

        Initiate exposures instance with value equal to the original BlackMarble
        nightlight intensity resampled to the target resolution `res_arcsec`.

        Provide either `countries` or `shape`.

        Parameters
        ----------
        countries : list or str, optional
            list containing country identifiers (name or iso3)
        shape : Shape, Polygon or MultiPolygon, optional
            geographical shape of target region, alternative to `countries`.
        res_arcsec : int, optional
            Resolution in arc seconds. The default is 15.
        reference_year : int, optional
            Reference year. The default is CONFIG.exposures.def_ref_year.
        data_dir : Path, optional
            data directory. The default is None.

        Raises
        ------
        ValueError

        Returns
        -------
        exp : LitPop
            Exposure instance with values representing pure nightlight intensity
            from input nightlight data (BlackMarble)
        """
        if countries is None and shape is None:
            raise ValueError("Either `countries` or `shape` required. Aborting.")
        if countries is not None and shape is not None:
            raise ValueError("Not allowed to set both `countries` and `shape`. Aborting.")
        if countries is not None:
            exp = cls.from_countries(countries, res_arcsec=res_arcsec,
                                     exponents=(1,0), fin_mode='none',
                                     reference_year=reference_year, gpw_version=GPW_VERSION,
                                     data_dir=data_dir)
        else:
            exp = cls.from_shape(shape, None, res_arcsec=res_arcsec,
                                 exponents=(1,0), value_unit='',
                                 reference_year=reference_year,
                                 gpw_version=GPW_VERSION, data_dir=SYSTEM_DIR)
        LOGGER.warning("Note: set_nightlight_intensity sets values to raw nightlight intensity, "
                       "not to USD. "
                       "To disaggregate asset value proportionally to nightlights^m, "
                       "call from_countries or from_shape with exponents=(m,0).")
        return exp

    def set_population(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_population instead."""
        LOGGER.warning("The use of LitPop.set_population is deprecated."
                       "Use LitPop.from_population instead.")
        self.__dict__ = LitPop.from_population(*args, **kwargs).__dict__

    @classmethod
    def from_population(cls, countries=None, shape=None, res_arcsec=30,
                        reference_year=DEF_REF_YEAR, gpw_version=GPW_VERSION,
                        data_dir=SYSTEM_DIR):
        """
        Wrapper around `from_countries` / `from_shape`.

        Initiate exposures instance with value equal to GPW population count.
        Provide either `countries` or `shape`.

        Parameters
        ----------
        countries : list or str, optional
            list containing country identifiers (name or iso3)
        shape : Shape, Polygon or MultiPolygon, optional
            geographical shape of target region, alternative to `countries`.
        res_arcsec : int, optional
            Resolution in arc seconds. The default is 30.
        reference_year : int, optional
            Reference year (closest available GPW data year is used)
            The default is CONFIG.exposures.def_ref_year.
        gpw_version : int, optional
            specify GPW data verison. The default is 11.
        data_dir : Path, optional
            data directory. The default is None.
            Either countries or shape is required.

        Raises
        ------
        ValueError

        Returns
        -------
        exp : LitPop
            Exposure instance with values representing population count according
            to Gridded Population of the World (GPW) input data set.
        """
        if countries is None and shape is None:
            raise ValueError("Either `countries` or `shape` required. Aborting.")
        if countries is not None and shape is not None:
            raise ValueError("Not allowed to set both `countries` and `shape`. Aborting.")
        if countries is not None:
            exp = cls.from_countries(countries, res_arcsec=res_arcsec,
                                     exponents=(0,1), fin_mode='pop',
                                     reference_year=reference_year, gpw_version=gpw_version,
                                     data_dir=data_dir)
        else:
            exp = cls.from_shape(shape, None, res_arcsec=res_arcsec, exponents=(0,1),
                                 value_unit='people', reference_year=reference_year,
                                 gpw_version=gpw_version, data_dir=data_dir)
        return exp

    def set_custom_shape_from_countries(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_shape_and_countries instead."""
        LOGGER.warning("The use of LitPop.set_custom_shape_from_countries is deprecated."
                       "Use LitPop.from_shape_and_countries instead.")
        self.__dict__ = LitPop.from_shape_and_countries(*args, **kwargs).__dict__

    @classmethod
    def from_shape_and_countries(cls, shape, countries, res_arcsec=30, exponents=(1,1),
                                 fin_mode='pc', admin1_calc=False, reference_year=DEF_REF_YEAR,
                                 gpw_version=GPW_VERSION, data_dir=SYSTEM_DIR):
        """
        create LitPop exposure for `country` and then crop to given shape.

        Parameters
        ----------
        shape : shapely.geometry.Polygon, MultiPolygon, shapereader.Shape,
            or GeoSeries or list containg either Polygons or Multipolygons.
            Geographical shape for which LitPop Exposure is to be initiated.
        countries : list with str or int
            list containing country identifiers:
            iso3alpha (e.g. 'JPN'), iso3num (e.g. 92) or name (e.g. 'Togo')
        res_arcsec : float, optional
            Horizontal resolution in arc-sec.
            The default is 30 arcsec, this corresponds to roughly 1 km.
        exponents : tuple of two integers, optional
            Defining power with which lit (nightlights) and pop (gpw) go into LitPop.
            Default: (1, 1)
        fin_mode : str, optional
            Socio-economic value to be used as an asset base that is disaggregated
            to the grid points within the country
                * 'pc': produced capital (Source: World Bank), incl. manufactured or
                  built assets such as machinery, equipment, and physical structures
                  (pc is in constant 2014 USD)
                * 'pop': population count (source: GPW, same as gridded population).
                  The unit is 'people'.
                * 'gdp': gross-domestic product (Source: World Bank) [USD]
                * 'income_group': gdp multiplied by country's income group+1 [USD]
                  Income groups are 1 (low) to 4 (high income).
                * 'nfw': non-financial wealth (Source: Credit Suisse, of households only) [USD]
                * 'tw': total wealth (Source: Credit Suisse, of households only) [USD]
                * 'norm': normalized by country
                * 'none': LitPop per pixel is returned unchanged
            The default is 'pc'.
        admin1_calc : boolean, optional
            If True, distribute admin1-level GDP (if available). Default: False
        reference_year : int, optional
            Reference year for data sources. Default: 2020
        gpw_version : int, optional
            Version number of GPW population data.
            The default is GPW_VERSION
        data_dir : Path, optional
            redefines path to input data directory. The default is SYSTEM_DIR.

        Raises
        ------
        NotImplementedError

        Returns
        -------
        exp : LitPop
            The exposure LitPop within shape
        """
        # init countries' exposure:
        exp = cls.from_countries(countries, res_arcsec=res_arcsec, exponents=exponents,
                                fin_mode=fin_mode, reference_year=reference_year,
                                gpw_version=gpw_version, data_dir=data_dir)

        if isinstance(shape, Shape):
            # get gdf with geometries of points within shape:
            shape_gdf, _ = _get_litpop_single_polygon(shape, reference_year,
                                                      res_arcsec, data_dir,
                                                      gpw_version, exponents,
                                                      )
            shape_gdf = shape_gdf.drop(
                columns=shape_gdf.columns[shape_gdf.columns != 'geometry'])
            # extract gdf with data points within shape:
            gdf = geopandas.sjoin(exp.gdf, shape_gdf, how='right')
            gdf = gdf.drop(columns=['index_left'])
        elif isinstance(shape, (shapely.geometry.MultiPolygon, shapely.geometry.Polygon)):
            # works if shape is Polygon or MultiPolygon
            gdf = exp.gdf.loc[exp.gdf.geometry.within(shape)]
        elif isinstance(shape, (geopandas.GeoSeries, list)):
            gdf = geopandas.GeoDataFrame(columns=exp.gdf.columns)
            for shp in shape:
                if isinstance(shp, (shapely.geometry.MultiPolygon,
                                    shapely.geometry.Polygon)):
                    gdf = gdf.append(exp.gdf.loc[exp.gdf.geometry.within(shp)])
                else:
                    raise NotImplementedError('Not implemented for list or GeoSeries containing '
                                              f'objects of type {type(shp)} as `shape`')
        else:
            raise NotImplementedError('Not implemented for `shape` of type {type(shape)}')

        tag = Tag()
        tag.description = f'LitPop Exposure for custom shape in {countries} at ' \
                          f'{res_arcsec} as, year: {reference_year}, financial mode: ' \
                          f'{fin_mode}, exp: {exponents}, admin1_calc: {admin1_calc}'
        exp.gdf = gdf.reset_index()

        try:
            rows, cols, ras_trans = u_coord.pts_to_raster_meta(
                (exp.gdf.longitude.min(), exp.gdf.latitude.min(),
                 exp.gdf.longitude.max(), exp.gdf.latitude.max()),
                u_coord.get_resolution(exp.gdf.longitude, exp.gdf.latitude))
            exp.meta = {
                'width': cols,
                'height': rows,
                'crs': exp.crs,
                'transform': ras_trans,
            }
        except ValueError as err:
            LOGGER.warning('Could not write attribute meta with ValueError: ')
            LOGGER.warning(err.args[0])
            exp.meta = {'crs': exp.crs}
        return exp

    def set_custom_shape(self, *args, **kwargs):
        """This function is deprecated, use LitPop.from_shape instead."""
        LOGGER.warning("The use of LitPop.set_custom_shape is deprecated."
                       "Use LitPop.from_shape instead.")
        self.__dict__ = LitPop.from_shape(*args, **kwargs).__dict__

    @classmethod
    def from_shape(cls, shape, total_value, res_arcsec=30, exponents=(1,1), value_unit='USD',
                   reference_year=DEF_REF_YEAR, gpw_version=GPW_VERSION, data_dir=SYSTEM_DIR):
        """init LitPop exposure object for a custom shape.
        Requires user input regarding the total value to be disaggregated.

        Sets attributes `ref_year`, `tag`, `crs`, `value`, `geometry`, `meta`,
        `value_unit`, `exponents`,`fin_mode`, `gpw_version`, and `admin1_calc`.

        This method can be used to initiated LitPop Exposure for sub-national
        regions such as states, districts, cantons, cities, ... but shapes
        and total value need to be provided manually.
        If these required input parameters are not known / available,
        better initiate Exposure for entire country and extract shape afterwards.

        Parameters
        ----------
        shape : shapely.geometry.Polygon or MultiPolygon or shapereader.Shape.
            Geographical shape for which LitPop Exposure is to be initiated.
        total_value : int, float or None type
            Total value to be disaggregated to grid in shape.
            If None, no value is disaggregated.
        res_arcsec : float, optional
            Horizontal resolution in arc-sec.
            The default 30 arcsec corresponds to roughly 1 km.
        exponents : tuple of two integers, optional
            Defining power with which lit (nightlights) and pop (gpw) go into LitPop.
        value_unit : str
            Unit of exposure values. The default is USD.
        reference_year : int, optional
            Reference year for data sources. Default: CONFIG.exposures.def_ref_year
        gpw_version : int, optional
            Version number of GPW population data.
            The default is set in CONFIG.
        data_dir : Path, optional
            redefines path to input data directory. The default is SYSTEM_DIR.

        Raises
        ------
        NotImplementedError
        ValueError
        TypeError

        Returns
        -------
        exp : LitPop
            The exposure LitPop within shape
        """
        if isinstance(shape, (geopandas.GeoSeries, list)):
            raise NotImplementedError('Not implemented for `shape` of type list or '
                                      'GeoSeries. Loop over elements of series outside method.')

        tag = Tag()
        litpop_gdf, _ = _get_litpop_single_polygon(shape, reference_year,
                                                          res_arcsec, data_dir,
                                                          gpw_version, exponents,
                                                          )

        # disaggregate total value proportional to LitPop values:
        if isinstance(total_value, (float, int)):
            litpop_gdf['value'] = np.divide(litpop_gdf['value'],
                                            litpop_gdf['value'].sum()) * total_value
        elif total_value is not None:
            raise TypeError("total_value must be int, float or None.")

        tag.description = f'LitPop Exposure for custom shape at {res_arcsec} as, ' \
                          f'year: {reference_year}, exp: {exponents}'


        litpop_gdf[INDICATOR_IMPF] = 1

        exp = cls(
                  data=litpop_gdf,
                  crs=litpop_gdf.crs,
                  ref_year=reference_year,
                  tag=tag,
                  value_unit=value_unit,
                  exponents = exponents,
                  gpw_version = gpw_version,
                  fin_mode = None,
                  )

        if min(len(exp.gdf.latitude.unique()), len(exp.gdf.longitude.unique())) > 1:
        #if exp.gdf.shape[0] > 1 and len(exp.gdf.latitude.unique()) > 1:
            rows, cols, ras_trans = u_coord.pts_to_raster_meta(
                (exp.gdf.longitude.min(), exp.gdf.latitude.min(),
                 exp.gdf.longitude.max(), exp.gdf.latitude.max()),
                u_coord.get_resolution(exp.gdf.longitude, exp.gdf.latitude))
            exp.meta = {
                'width': cols,
                'height': rows,
                'crs': exp.crs,
                'transform': ras_trans,
            }
        else:
            LOGGER.warning('Could not write attribute meta because coordinates'
                           'are either only one point or do not extend in lat and lon')
            exp.meta = {'crs': exp.crs}
        return exp

    @staticmethod
    def _from_country(country, res_arcsec=30, exponents=(1,1), fin_mode=None,
                         total_value=None, reference_year=DEF_REF_YEAR,
                         gpw_version=GPW_VERSION, data_dir=SYSTEM_DIR):
        """init LitPop exposure object for one single country
        See docstring of from_countries() for detailled description of parameters.

        Parameters
        ----------
        country : str or int
            country identifier as iso3alpha, iso3num or name.
        res_arcsec : float, optional
            horizontal resolution in arc-sec.
        exponents : tuple of two integers, optional
        fin_mode : str, optional
        total_value : numeric, optional
        reference_year : int, optional
        gpw_version : int, optional
        data_dir : Path, optional
            redefines path to input data directory. The default is SYSTEM_DIR.

        Raises
        ------
        ValueError

        Returns
        -------
        exp : LitPop
            LitPop Exposure instance for the country
        """
        # Determine ISO 3166 representation of country and get geometry:
        try:
            iso3a = u_coord.country_to_iso(country, representation="alpha3")
            iso3n = u_coord.country_to_iso(country, representation="numeric")
        except LookupError:
            LOGGER.error('Country not identified: %s.', country)
            return None
        country_geometry = u_coord.get_land_geometry([iso3a])
        if not country_geometry.bounds: # check for empty shape
            LOGGER.error('No geometry found for country: %s.', country)
            return None
        LOGGER.info('\n LitPop: Init Exposure for country: %s (%i)...\n',
                    iso3a, iso3n)
        litpop_gdf = geopandas.GeoDataFrame()
        total_population = 0

        # for countries with multiple sperated shapes (e.g., islands), data
        # is initiated for each shape separately and 0 values (e.g. on sea)
        # removed before combination, to save memory.
        # loop over single polygons in country shape object:
        for idx, polygon in enumerate(list(country_geometry)):
            # get litpop data for each polygon and combine into GeoDataFrame:
            gdf_tmp, meta_tmp, = \
                _get_litpop_single_polygon(polygon, reference_year,
                                           res_arcsec, data_dir,
                                           gpw_version, exponents,
                                           verbatim=not bool(idx),
                                           region_id=iso3n
                                           )
            if gdf_tmp is None:
                LOGGER.debug(f'Skipping polygon with index {idx} for' +
                               f' country {iso3a}.')
                continue
            total_population += meta_tmp['total_population']
            litpop_gdf = litpop_gdf.append(gdf_tmp)
        litpop_gdf.crs = meta_tmp['crs']

        # set total value for disaggregation if not provided:
        if total_value is None and fin_mode == 'pop':
            total_value = total_population # population count is taken from pop-data.
        elif total_value is None:
            total_value = _get_total_value_per_country(iso3a, fin_mode, reference_year)

        # disaggregate total value proportional to LitPop values:
        if isinstance(total_value, (float, int)):
            litpop_gdf['value'] = np.divide(litpop_gdf['value'],
                                            litpop_gdf['value'].sum()) * total_value
        elif total_value is not None:
            raise TypeError("total_value must be int or float.")

        exp = LitPop()
        exp.set_gdf(litpop_gdf)
        exp.gdf[INDICATOR_IMPF] = 1
        return exp

    # Alias method names for backward compatibility:
    set_country = set_countries

def _get_litpop_single_polygon(polygon, reference_year, res_arcsec, data_dir,
                               gpw_version, exponents, region_id=None, verbatim=False):
    """load nightlight (nl) and population (pop) data in rastered 2d arrays
    and apply rescaling (resolution reprojection) and LitPop core calculation,
    i.e. combination of nl and pop per grid cell.

    NB: In this method, there is no disaggregation taking place, yet. This is
    done in a subsequent step outside this method.

    Parameters
    ----------
    polygon : Polygon object
        single polygon to be extracted
    reference_year : int, optional
        Reference year for data sources.
    res_arcsec : float
        Horizontal resolution in arc-seconds.
    data_dir : Path
        redefines path to input data directory. The default is SYSTEM_DIR.
    gpw_version : int
        Version number of GPW population data.
        The default is None. If None, the default is set in gpw_population module.
    exponents : tuple of two integers
        Defining power with which lit (nightlights) and pop (gpw) go into LitPop. To get
        nightlights^3 without population count: (3, 0). To use population count alone: (0, 1).
    region_id : int, optional
        if provided, region_id of gdf is set to value.
        The default is None, this implies that region_id is not set.
    verbatim : bool, optional
        verbatim logging? Default is False.

    Returns
    -------
    litpop_gdf : GeoDataframe
        resulting gridded data for Lit^m * Pop^n inside polygon,
        data points outside the polygon and equal to zero are not returned.
    meta_out : dict
        raster meta info for gridded data in litpop_array, additionally the field
        'total_population' contains the sum of population in the polygon.

    """
    # set nightlight offset (delta) to 1 in case n>0, c.f. delta in Eq. 1 of paper:
    if exponents[1] == 0:
        offsets = (0, 0)
    else:
        offsets = (1, 0)
    # import population data (2d array), meta data, and global grid info,
    # global_transform defines the origin (corner points) of the global traget grid:
    pop, meta_pop, global_transform = \
        pop_util.load_gpw_pop_shape(polygon,
                                    reference_year,
                                    gpw_version=gpw_version,
                                    data_dir=data_dir,
                                    verbatim=verbatim,
                                    )
    total_population = pop.sum()
    # import nightlight data (2d array) and associated meta data:
    nlight, meta_nl = nl_util.load_nasa_nl_shape(polygon,
                                                 reference_year,
                                                 data_dir=data_dir,
                                                 dtype=float
                                                 )

    # if resolution is the same as for lit (15 arcsec), set grid same as lit:
    if res_arcsec==15:
        i_align = 1
        global_origins = (meta_nl['transform'][2], # lon
                          meta_nl['transform'][5]) # lat
    else: # align grid for resampling to grid of population data (pop)
        i_align = 0
        global_origins=(global_transform[2],
                        global_transform[5])

    # reproject Lit and Pop input data to aligned grid with target resolution:
    try:
        [pop, nlight], meta_out = reproject_input_data([pop, nlight],
                                                  [meta_pop, meta_nl],
                                                  i_align=i_align, # pop defines grid
                                                  target_res_arcsec=res_arcsec,
                                                  global_origins=global_origins,
                                                  )
    except ValueError as err:
        if "height must be > 0" in err.args[0] or "width must be > 0" in err.args[0]:
            # no grid point within shape after reprojection, None is returned.
            if verbatim:
                LOGGER.info('No data point on destination grid within polygon.')
            return None, {'crs': meta_pop['crs']}
        raise err

    # calculate Lit^m * Pop^n (but not yet disaggregate any total value to grid):
    litpop_array = gridpoints_core_calc([nlight, pop],
                                        offsets=offsets,
                                        exponents=exponents,
                                        total_val_rescale=None)

    # mask entries outside polygon (set to NaN) and set total population:
    litpop_array = u_coord.mask_raster_with_geometry(litpop_array, meta_out['transform'],
                                                     [polygon], nodata=np.nan)
    meta_out['total_population'] = total_population

    # extract coordinates as meshgrid arrays:
    lon, lat = u_coord.raster_to_meshgrid(meta_out['transform'],
                                          meta_out['width'],
                                          meta_out['height'])
    # init GeoDataFrame from data and coordinates:
    gdf = geopandas.GeoDataFrame({'value': litpop_array.flatten()}, crs=meta_out['crs'],
                                 geometry=geopandas.points_from_xy(
                                     np.round_(lon.flatten(), decimals = 8, out = None),
                                     np.round_(lat.flatten(), decimals = 8, out = None)
                                     )
                                 )
    gdf['latitude'] = np.round_(lat.flatten(), decimals = 8, out = None)
    gdf['longitude'] = np.round_(lon.flatten(), decimals = 8, out = None)
    if region_id is not None: # set region_id
        gdf['region_id'] = region_id
    else:
        gdf['region_id'] = u_coord.get_country_code(gdf.latitude, gdf.longitude)
    # remove entries outside polygon with `dropna` and return GeoDataFrame:
    return gdf.dropna(), meta_out

def get_value_unit(fin_mode):
    """get `value_unit` depending on `fin_mode`

    Parameters
    ----------
    fin_mode : Socio-economic value to be used as an asset base

    Returns
    -------
    value_unit : str

    """
    if fin_mode in ['none', 'norm']:
        return ''
    if fin_mode == 'pop':
        return 'people'
    return 'USD'

def _get_total_value_per_country(cntry_iso3a, fin_mode, reference_year):
    """
    Get total value for disaggregation, e.g., total asset value
    for a country, depending on user choice (fin_mode).

    Parameters
    ----------
    cntry_iso3a : str
        country iso3 code alphabetic, e.g. 'JPN' for Japan
    fin_mode : str
        Socio-economic value to be used as an asset base that is disaggregated
        to the grid points within the country
            * 'pc': produced capital (Source: World Bank), incl. manufactured or
              built assets such as machinery, equipment, and physical structures
              (pc is in constant 2014 USD)
            * 'pc_land': produced capital (Source: World Bank), incl. manufactured or
              built assets such as machinery, equipment, physical structures,
              and land value for built-up land.
              (pc is in constant 2014 USD)
            * 'gdp': gross-domestic product (Source: World Bank) [USD]
            * 'income_group': gdp multiplied by country's income group+1 [USD]
              Income groups are 1 (low) to 4 (high income).
            * 'nfw': non-financial wealth (Source: Credit Suisse, of households only) [USD]
            * 'tw': total wealth (Source: Credit Suisse, of households only) [USD]
            * 'norm': normalized by country
            * 'none': LitPop per pixel is returned unscaled
        The default is 'pc'
    reference_year : int
        reference year for data extraction

    Returns
    -------
    total_value : float
    """
    if fin_mode == 'none':
        return None
    if fin_mode == 'pop':
        raise NotImplementedError("`_get_total_value_per_country` is not "
                                  "implemented for `fin_mode` == 'pop'.")
    if fin_mode == 'pc':
        return(u_fin.world_bank_wealth_account(cntry_iso3a, reference_year,
                                               no_land=True)[1])
        # here, total_asset_val is Produced Capital "pc"
        # no_land=True returns value w/o the mark-up of 24% for land value
    if fin_mode == 'pc_land':
        return(u_fin.world_bank_wealth_account(cntry_iso3a, reference_year,
                                               no_land=False)[1])
        # no_land=False returns pc value incl. the mark-up of 24% for land value
    if fin_mode == 'norm':
        return 1
    # GDP based total values:
    gdp_value = u_fin.gdp(cntry_iso3a, reference_year)[1]
    if fin_mode == 'gdp':
        return gdp_value
    if fin_mode == 'income_group': # gdp * (income group + 1)
        return gdp_value*(u_fin.income_group(cntry_iso3a, reference_year)[1]+1)
    if fin_mode in ('nfw', 'tw'):
        wealthtogdp_factor = u_fin.wealth2gdp(cntry_iso3a, fin_mode == 'nfw',
                                              reference_year)[1]
        if np.isnan(wealthtogdp_factor):
            LOGGER.warning("Missing wealth-to-gdp factor for country %s.", cntry_iso3a)
            LOGGER.warning("Using GDP instead as total value.")
            return gdp_value
        return gdp_value * wealthtogdp_factor
    raise ValueError(f"Unsupported fin_mode: {fin_mode}")

def reproject_input_data(data_array_list, meta_list,
                        i_align=0,
                        target_res_arcsec=None,
                        global_origins=(-180.0, 89.99999999999991),
                        resampling=rasterio.warp.Resampling.bilinear,
                        conserve=None):
    """
    LitPop-sepcific wrapper around u_coord.align_raster_data.

    Reprojects all arrays in data_arrays to a given resolution –
    all based on the population data grid.

    Parameters
    ----------
    data_array_list : list or array of numpy arrays containing numbers
        Data to be reprojected, i.e. list containing N (min. 1) 2D-arrays.
        The data with the reference grid used to align the global destination
        grid to should be first data_array_list[i_align],
        e.g., pop (GPW population data) for LitPop.
    meta_list : list of dicts
        meta data dictionaries of data arrays in same order as data_array_list.
        Required fields in each dict are 'dtype,', 'width', 'height', 'crs', 'transform'.
        Example:
            {'driver': 'GTiff',
             'dtype': 'float32',
             'nodata': 0,
             'width': 2702,
             'height': 1939,
             'count': 1,
             'crs': CRS.from_epsg(4326),
             'transform': Affine(0.00833333333333333, 0.0, -18.175000000000068,
                                 0.0, -0.00833333333333333, 43.79999999999993)}
        The meta data with the reference grid used to define the global destination
        grid should be first in the list, e.g., GPW population data for LitPop.
    i_align : int, optional
        Index/Position of meta in meta_list to which the global grid of the destination
        is to be aligned to (c.f. u_coord.align_raster_data)
        The default is 0.
    target_res_arcsec : int, optional
        target resolution in arcsec. The default is None, i.e. same resolution
        as reference data.
    global_origins : tuple with two numbers (lat, lon), optional
        global lon and lat origins as basis for destination grid.
        The default is the same as for GPW population data:
            (-180.0, 89.99999999999991)
    resampling : resampling function, optional
        The default is rasterio.warp.Resampling.bilinear
    conserve : str, optional, either 'mean' or 'sum'
        Conserve mean or sum of data? The default is None (no conservation).

    Returns
    -------
    data_array_list : list
        contains reprojected data sets
    meta_out : dict
        contains meta data of new grid (same for all arrays)
    """

    # target resolution in degree lon,lat:
    if target_res_arcsec is None:
        res_degree = meta_list[i_align]['transform'][0] # reference grid
    else:
        res_degree = target_res_arcsec / 3600

    dst_crs = meta_list[i_align]['crs']
    # loop over data arrays, do transformation where required:
    data_out_list = [None] * len(data_array_list)
    meta_out = {'dtype': meta_list[i_align]['dtype'],
                'nodata': meta_list[i_align]['dtype'],
                'crs': dst_crs}

    for idx, data in enumerate(data_array_list):
        # if target resolution corresponds to reference data resolution,
        # the reference data is not transformed:
        if idx==i_align and ((target_res_arcsec is None) or \
                           (np.round(meta_list[i_align]['transform'][0],
                            decimals=7)==np.round(res_degree, decimals=7))):
            data_out_list[idx] = data
            continue
        # reproject data grid:
        dst_bounds = rasterio.transform.array_bounds(meta_list[i_align]['height'],
                                                     meta_list[i_align]['width'],
                                                     meta_list[i_align]['transform'])
        data_out_list[idx], meta_out['transform'] = \
            u_coord.align_raster_data(data_array_list[idx], meta_list[idx]['crs'],
                                      meta_list[idx]['transform'],
                                      dst_crs=dst_crs,
                                      dst_resolution=(res_degree, res_degree),
                                      dst_bounds=dst_bounds,
                                      global_origin=global_origins,
                                      resampling=resampling,
                                      conserve=conserve)
    meta_out['height'] = data_out_list[-1].shape[0]
    meta_out['width'] = data_out_list[-1].shape[1]
    return data_out_list, meta_out

def gridpoints_core_calc(data_arrays, offsets=None, exponents=None,
                         total_val_rescale=None):
    """
    Combines N dense numerical arrays by point-wise multipilcation and
    optionally rescales to new total value:
    (1) An offset (1 number per array) is added to all elements in
        the corresponding data array in data_arrays (optional).
    (2) Numbers in each array are taken to the power of the corresponding
        exponent (optional).
    (3) Arrays are multiplied element-wise.
    (4) if total_val_rescale is provided,
        results are normalized and re-scaled with total_val_rescale.
    (5) One array with results is returned.

    Parameters
    ----------
    data_arrays : list or array of numpy arrays containing numbers
        Data to be combined, i.e. list containing N (min. 1) arrays of same shape.
    total_val_rescale : float or int, optional
        Total value for optional rescaling of resulting array. All values in result_array
        are skaled so that the sum is equal to total_val_rescale.
        The default (None) implies no rescaling.
    offsets: list or array containing N numbers >= 0, optional
        One numerical offset per array that is added (sum) to the
        corresponding array in data_arrays.
        The default (None) corresponds to np.zeros(N).
    exponents: list or array containing N numbers >= 0, optional
        One exponent per array used as power for the corresponding array.
        The default (None) corresponds to np.ones(N).

    Raises
    ------
    ValueError
        If input lists don't have the same number of elements.
        Or: If arrays in data_arrays do not have the same shape.

    Returns
    -------
    result_array : np.array of same shape as arrays in data_arrays
        Results from calculation described above.
    """
    # convert input data to arrays
    # check integrity of data_array input (length and type):
    try:
        data_arrays = [np.array(data) for data in data_arrays]
        if len(list({data.shape for data in data_arrays})) > 1:
            raise ValueError("Elements in data_arrays don't agree in shape.")
    except AttributeError as err:
        raise TypeError("data_arrays or contained elements have wrong type.") from err

    # if None, defaults for offsets and exponents are set:
    if offsets is None:
        offsets = np.zeros(len(data_arrays))
    if exponents is None:
        exponents = np.ones(len(data_arrays))
    if np.min(offsets) < 0:
        raise ValueError("offset values < 0 are not allowed, because negative offset "
                         "values produce undesired negative disaggregation values.")
    if np.min(exponents) < 0:
        raise ValueError("exponents < 0 are not allowed, because negative exponents "
                         "produce 'inf' values where the initial value is 0.")
    # Steps 1-3: arrays are multiplied after application of offets and exponents:
    #       (arrays are converted to float to prevent ValueError:
    #       "Integers to negative integer powers are not allowed.")
    result_array = np.power(np.array(data_arrays[0]+offsets[0], dtype=float),
                            exponents[0])
    for i in np.arange(1, len(data_arrays)):
        result_array = np.multiply(result_array,
                                   np.power(np.array(data_arrays[i]+offsets[i], dtype=float),
                                            exponents[i])
                                   )

    # Steps 4+5: if total value for rescaling is provided, result_array is normalized and
    # scaled with this total value (total_val_rescale):
    if isinstance(total_val_rescale, (float, int)):
        return np.divide(result_array, result_array.sum()) * total_val_rescale
    if total_val_rescale is not None:
        raise TypeError("total_val_rescale must be int or float.")
    return result_array

# The following functions are only required if calc_admin1 is True,
# not for core LitPop. They are maintained here mainly for backward compatibility
def _check_excel_exists(file_path, file_name, xlsx_before_xls=True):
    """Checks if an Excel file with the name file_name in the folder file_path exists, checking for
    both xlsx and xls files.

    Parameters
    ----------
    file_path : string
        path where to look for file
    file_name : string
        file name which is checked. Extension is ignored
    xlsx_before_xls : boolean, optional
        If True, xlsx files are priorised over xls files. Default: True.

    Returns
    -------
    path: Path instance
    """
    try_ext = []
    if xlsx_before_xls:
        try_ext.append('.xlsx')
        try_ext.append('.xls')
    else:
        try_ext.append('.xls')
        try_ext.append('.xlsx')
    path_name = Path(file_path, file_name).stem
    for i in try_ext:
        if Path(file_path, path_name + i).is_file():
            return str(Path(file_path, path_name + i))
    return None

def _grp_read(country_iso3, admin1_info=None, data_dir=SYSTEM_DIR):
    """Retrieves the Gross Regional Product (GRP) aka Gross State Domestic Product (GSDP)
    data for a certain country. It requires an excel file in a subfolder
    "GSDP" in climadas data folder (or in the specified folder). The excel file should bear the
    name 'ISO3_GSDP.xlsx' (or .xls), where ISO3 is the three letter country code. In the excel
    file, the first sheet should contain a row with the title "State_Province" with the name or
    postal code (two letters) of the admin1 unit and a row "GSDP_ref" with either the GDP value of
    the admin1 unit or its share in the national GDP.

    Parameters
    ----------
    country_iso3 : str
        alphabetic three letter country code ISO3a
    admin1_info : list, optional
        list containg all admin1 records for country.
        if not provided, info is set retrieved automatically
    data_dir : str, optional
        path where to look for file

    Returns
    -------
    out_dict : dictionary
        GRP for each admin1 unit name.
    """
    if admin1_info is None:
        admin1_info, _ = u_coord.get_admin1_info(country_iso3)
        admin1_info = admin1_info[country_iso3]
    file_name = _check_excel_exists(data_dir.joinpath('GSDP'), str(country_iso3 + '_GSDP'))
    if file_name is not None:
        # open spreadsheet and identify relevant columns:
        admin1_xls_data = pd.read_excel(file_name)
        if admin1_xls_data.get('State_Province') is None:
            admin1_xls_data = admin1_xls_data.rename(
                columns={admin1_xls_data.columns[0]: 'State_Province'})
        if admin1_xls_data.get('GSDP_ref') is None:
            admin1_xls_data = admin1_xls_data.rename(
                columns={admin1_xls_data.columns[-1]: 'GSDP_ref'})

        # initiate dictionary with admin 1 names as keys:
        out_dict = dict.fromkeys([record['name'] for record in admin1_info])
        postals = [record['postal'] for record in admin1_info]
        # first nested loop. outer loop over region names in admin1_info:
        for record_name in out_dict:
            # inner loop over region names in spreadsheet, find matches
            for idx, xls_name in enumerate(admin1_xls_data['State_Province'].tolist()):
                subnat_shape_str = [c for c in record_name if c.isalpha() or c.isnumeric()]
                subnat_xls_str = [c for c in xls_name if c.isalpha()]
                if subnat_shape_str == subnat_xls_str:
                    out_dict[record_name] = admin1_xls_data['GSDP_ref'][idx]
                    break
        # second nested loop to detect matched empty entries
        for idx1, country_name in enumerate(out_dict.keys()):
            if out_dict[country_name] is None:
                for idx2, xls_name in enumerate(admin1_xls_data['State_Province'].tolist()):
                    subnat_xls_str = [c for c in xls_name if c.isalpha()]
                    postals_str = [c for c in postals[idx1] if c.isalpha()]
                    if subnat_xls_str == postals_str:
                        out_dict[country_name] = admin1_xls_data['GSDP_ref'][idx2]
        return out_dict
    LOGGER.warning('No file for %s could be found in %s.', country_iso3, data_dir)
    LOGGER.warning('No admin1 data is calculated in this case.')
    return None

def _calc_admin1_one_country(country, res_arcsec, exponents, fin_mode, total_value,
                 reference_year, gpw_version, data_dir):
    """
    Calculates the LitPop on admin1 level for provinces/states where such information are
    available (i.e. GDP is distributed on a subnational instead of a national level). Requires
    excel files in a subfolder "GSDP" in climadas data folder. The excel files should contain a row
    with the title "State_Province" with the name or postal code (two letters) of the admin1 unit
    and a row "GSDP_ref" with either the GDP value or the share of the state in the national GDP.
    If only for certain states admin1 info is found, the rest of the country is assigned value
    according to the admin0 method.

    See from_countries() for description of parameters.

    Parameters
    ----------
    country : str
    res_arcsec : int
    exponents : tuple
    fin_mode : str
    total_value :int or float
    reference_year : int
    gpw_version: int
    data_dir : Path

    Returns
    -------
    Exposure instance

    """
    if fin_mode == 'pop':
        raise NotImplementedError('`_calc_admin1_one_country` not implemented for '+
                                          "`fin_mode` == 'pop'.")
    # Determine ISO 3166 representation of country and get geometry:
    try:
        iso3a = u_coord.country_to_iso(country, representation="alpha3")
    except LookupError:
        LOGGER.error('Country not identified: %s. Skippig.', country)
        return None
    # get records and shapes on admin 1 level:
    admin1_info, admin1_shapes = u_coord.get_admin1_info(iso3a)
    admin1_info = admin1_info[iso3a]
    admin1_shapes = admin1_shapes[iso3a]
    # get subnational Gross Regional Product (GRP) data for country:
    grp_values = _grp_read(iso3a, admin1_info=admin1_info, data_dir=data_dir)
    if grp_values is None:
        LOGGER.error("No subnational GRP data found for calc_admin1"
                         " for country %s. Skipping.", country)
        return None
    # normalize GRP values:
    sum_vals = sum(filter(None, grp_values.values())) # get total
    grp_values = {key: (value / sum_vals if value is not None else None)
                 for (key, value) in grp_values.items()}

    # get total value of country:
    total_value = _get_total_value_per_country(iso3a, fin_mode, reference_year)
    exp_list = []
    for idx, record in enumerate(admin1_info):
        if grp_values[record['name']] is None:
            continue
        LOGGER.info(record['name'])
        # init exposure for province and add to list
        # total value is defined from country multiplied by grp_share:
        exp_list.append(LitPop.from_shape(admin1_shapes[idx],
                                          total_value * grp_values[record['name']],
                                          res_arcsec=res_arcsec,
                                          exponents=exponents,
                                          reference_year=reference_year,
                                          gpw_version=gpw_version,
                                          data_dir=data_dir)
                        )
        exp_list[-1].gdf['admin1'] = record['name']

    return Exposures.concat(exp_list)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test hdf5_handler module.
"""

import unittest
import numpy as np
import h5py

from climada.util.constants import HAZ_DEMO_MAT
import climada.util.hdf5_handler as u_hdf5

class TestFunc(unittest.TestCase):
    """Test the auxiliary functions used to retrieve variables from HDF5"""

    def test_get_string_pass(self):
        """Check function to get a string from input integer array"""

        # Load input
        contents = u_hdf5.read(HAZ_DEMO_MAT)

        # Convert several strings
        str_date = u_hdf5.get_string(contents['hazard']['date'])
        str_comment = u_hdf5.get_string(contents['hazard']['comment'])
        str_wf = u_hdf5.get_string(contents['hazard']['windfield_comment'])
        str_fn = u_hdf5.get_string(contents['hazard']['filename'])

        # Check results
        self.assertEqual('14-Nov-2017 10:09:05', str_date)
        self.assertEqual(
            'TC hazard event set, generated 14-Nov-2017 10:09:05',
            str_comment)
        self.assertEqual(
            'generating 14450 windfields took 0.25 min ' +
            '(0.0010 sec/event)', str_wf)
        self.assertEqual('/Users/aznarsig/Documents/MATLAB/climada_data/' +
                         'hazards/atl_prob.mat', str_fn)

    def test_get_sparse_mat_pass(self):
        """Check contents of imported sparse matrix, using the function \
        to build a sparse matrix from the read HDF5 variable"""

        # Load input
        contents = u_hdf5.read(HAZ_DEMO_MAT)

        # get matrix size
        mat_shape = (len(contents['hazard']['event_ID']),
                     len(contents['hazard']['centroid_ID']))
        spr_mat = u_hdf5.get_sparse_csr_mat(
            contents['hazard']['intensity'], mat_shape)

        self.assertEqual(mat_shape[0], spr_mat.shape[0])
        self.assertEqual(mat_shape[1], spr_mat.shape[1])
        self.assertEqual(0, spr_mat[0, 0])
        self.assertEqual(9.2029355562400745, spr_mat[7916, 98])
        self.assertEqual(34.426829019435729, spr_mat[12839, 96])
        self.assertEqual(61.81217342446773, spr_mat[1557, 97])
        self.assertEqual(6.4470644550625753, spr_mat[6658, 97])
        self.assertEqual(0, spr_mat[298, 9])
        self.assertEqual(19.385821399329995, spr_mat[15, 0])
        self.assertEqual(52.253444444444447, spr_mat[76, 95])
        self.assertEqual(0, spr_mat[126, 86])

    def test_get_str_from_ref(self):
        """Check import string from a HDF5 object reference"""
        file = h5py.File(HAZ_DEMO_MAT, 'r')
        var = file['hazard']['name'][0][0]
        res = u_hdf5.get_str_from_ref(HAZ_DEMO_MAT, var)
        self.assertEqual('NNN_1185101', res)

    def test_get_list_str_from_ref(self):
        """Check import string from a HDF5 object reference"""
        file = h5py.File(HAZ_DEMO_MAT, 'r')
        var = file['hazard']['name']
        var_list = u_hdf5.get_list_str_from_ref(HAZ_DEMO_MAT, var)
        self.assertEqual('NNN_1185101', var_list[0])
        self.assertEqual('NNN_1185101_gen1', var_list[1])
        self.assertEqual('NNN_1185101_gen2', var_list[2])

class TestReader(unittest.TestCase):
    """Test HDF5 reader"""

    def test_hazard_pass(self):
        """Checking result against matlab atl_prob.mat file"""

        # Load input
        contents = u_hdf5.read(HAZ_DEMO_MAT)

        # Check read contents
        self.assertEqual(1, len(contents))
        self.assertTrue('hazard' in contents.keys())
        self.assertEqual(False, '#refs#' in contents.keys())

        hazard = contents['hazard']
        self.assertTrue('reference_year' in hazard.keys())
        self.assertTrue('lon' in hazard.keys())
        self.assertTrue('lat' in hazard.keys())
        self.assertTrue('centroid_ID' in hazard.keys())
        self.assertTrue('orig_years' in hazard.keys())
        self.assertTrue('orig_event_count' in hazard.keys())
        self.assertTrue('event_count' in hazard.keys())
        self.assertTrue('event_ID' in hazard.keys())
        self.assertTrue('category' in hazard.keys())
        self.assertTrue('orig_event_flag' in hazard.keys())
        self.assertTrue('yyyy' in hazard.keys())
        self.assertTrue('mm' in hazard.keys())
        self.assertTrue('dd' in hazard.keys())
        self.assertTrue('datenum' in hazard.keys())
        self.assertTrue('scenario' in hazard.keys())
        self.assertTrue('intensity' in hazard.keys())
        self.assertTrue('name' in hazard.keys())
        self.assertTrue('frequency' in hazard.keys())
        self.assertTrue('matrix_density' in hazard.keys())
        self.assertTrue('windfield_comment' in hazard.keys())
        self.assertTrue('peril_ID' in hazard.keys())
        self.assertTrue('filename' in hazard.keys())
        self.assertTrue('comment' in hazard.keys())
        self.assertTrue('date' in hazard.keys())
        self.assertTrue('units' in hazard.keys())
        self.assertTrue('orig_yearset' in hazard.keys())
        self.assertTrue('fraction' in hazard.keys())
        self.assertEqual(27, len(hazard.keys()))

        # Check some random values
        mat_shape = (len(contents['hazard']['event_ID']),
                     len(contents['hazard']['centroid_ID']))
        sp_mat = u_hdf5.get_sparse_csr_mat(hazard['intensity'], mat_shape)

        self.assertTrue(np.array_equal(np.array([[84], [67]]),
                                       hazard['peril_ID']))
        self.assertEqual(34.537289477809473, sp_mat[2862, 97])
        self.assertEqual(-80, hazard['lon'][46])
        self.assertEqual(28, hazard['lat'][87])
        self.assertEqual(2016, hazard['reference_year'])

    def test_with_refs_pass(self):
        """Allow to load references of the matlab file"""

        # Load input
        refs = True
        contents = u_hdf5.read(HAZ_DEMO_MAT, refs)

        # Check read contents
        self.assertEqual(2, len(contents))
        self.assertTrue('hazard' in contents.keys())
        self.assertTrue('#refs#' in contents.keys())

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestReader)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFunc))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""A setuptools based setup module.
"""

from pathlib import Path
from setuptools import find_packages, setup

here = Path(__file__).parent.absolute()

# Get the long description from the README file
with open(here.joinpath('README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='climada',

    version='3.1.2',

    description='CLIMADA in Python',

    long_description=long_description,
    long_description_content_type="text/markdown",

    url='https://github.com/CLIMADA-project/climada_python',

    author='ETH',
    author_email='schmide@ethz.ch',

    license='OSI Approved :: GNU Lesser General Public License v3 (GPLv3)',

    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],

    keywords='climate adaptation',

    packages=find_packages(),

    install_requires=[
        'bottleneck',
        'cartopy',
        'cfgrib',
        'contextily',
        'dask',
        'deprecation',
        'geopandas',
        'h5py',
        'haversine',
        'matplotlib',
        'netcdf4',
        'numba',
        'overpy',
        'pandas',
        'pandas-datareader',
        'pathos',
        'peewee',
        'pillow',
        'pint',
        'pybufrkit',
        'pycountry',
        'rasterio',
        'scikit-learn',
        'statsmodels',
        'tables',
        'tabulate',
        'tqdm',
        'xarray',
        'xlrd',
        'xlsxwriter',
        'xmlrunner'
    ],

    include_package_data=True
)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test of dates_times module
"""
import datetime as dt
import unittest
import numpy as np

import climada.util.dates_times as u_dt

class TestDateString(unittest.TestCase):
    """Test date functions"""
    def test_date_to_str_pass(self):
        """Test _date_to_str function"""
        ordinal_date = dt.datetime.toordinal(dt.datetime(2018, 4, 6))
        self.assertEqual('2018-04-06', u_dt.date_to_str(ordinal_date))

        ordinal_date = [dt.datetime.toordinal(dt.datetime(2018, 4, 6)),
                        dt.datetime.toordinal(dt.datetime(2019, 1, 1))]
        self.assertEqual(['2018-04-06', '2019-01-01'], u_dt.date_to_str(ordinal_date))

    def test_str_to_date_pass(self):
        """Test _date_to_str function"""
        date = 730000
        self.assertEqual(u_dt.str_to_date(u_dt.date_to_str(date)), date)

        date = [640000, 730000]
        self.assertEqual(u_dt.str_to_date(u_dt.date_to_str(date)), date)

class TestDateNumpy(unittest.TestCase):
    """Test date functions for numpy datetime64 type"""
    def test_datetime64_to_ordinal(self):
        """Test _datetime64_to_ordinal"""
        date = np.datetime64('1999-12-26T06:00:00.000000000')
        ordinal = u_dt.datetime64_to_ordinal(date)
        self.assertEqual(u_dt.date_to_str(ordinal), '1999-12-26')

        date = [np.datetime64('1999-12-26T06:00:00.000000000'),
                np.datetime64('2000-12-26T06:00:00.000000000')]
        ordinal = u_dt.datetime64_to_ordinal(date)
        self.assertEqual(u_dt.date_to_str(ordinal[0]), '1999-12-26')
        self.assertEqual(u_dt.date_to_str(ordinal[1]), '2000-12-26')

    def test_last_year_pass(self):
        """Test last_year"""
        ordinal_date = [dt.datetime.toordinal(dt.datetime(2018, 4, 6)),
                        dt.datetime.toordinal(dt.datetime(1918, 4, 6)),
                        dt.datetime.toordinal(dt.datetime(2019, 1, 1))]
        self.assertEqual(u_dt.last_year(ordinal_date), 2019)
        self.assertEqual(u_dt.last_year(np.array(ordinal_date)), 2019)

    def test_first_year_pass(self):
        """Test last_year"""
        ordinal_date = [dt.datetime.toordinal(dt.datetime(2018, 4, 6)),
                        dt.datetime.toordinal(dt.datetime(1918, 4, 6)),
                        dt.datetime.toordinal(dt.datetime(2019, 1, 1))]
        self.assertEqual(u_dt.first_year(ordinal_date), 1918)
        self.assertEqual(u_dt.first_year(np.array(ordinal_date)), 1918)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestDateString)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDateNumpy))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define Centroids class.
"""

import copy
import logging
from pathlib import Path

import cartopy.crs as ccrs
import geopandas as gpd
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pyproj.crs import CRS
import rasterio
from rasterio.warp import Resampling
from scipy import sparse
from shapely.geometry.point import Point

from climada.util.constants import (DEF_CRS,
                                    ONE_LAT_KM,
                                    NATEARTH_CENTROIDS)
import climada.util.coordinates as u_coord
import climada.util.hdf5_handler as u_hdf5
import climada.util.plot as u_plot

__all__ = ['Centroids']

PROJ_CEA = CRS.from_user_input({'proj': 'cea'})

DEF_VAR_MAT = {
    'field_names': ['centroids', 'hazard'],
    'var_name': {
        'lat': 'lat',
        'lon': 'lon',
        'dist_coast': 'distance2coast_km',
        'admin0_name': 'admin0_name',
        'admin0_iso3': 'admin0_ISO3',
        'comment': 'comment',
        'region_id': 'NatId'
    }
}
"""MATLAB variable names"""

DEF_VAR_EXCEL = {
    'sheet_name': 'centroids',
    'col_name': {
        'region_id': 'region_id',
        'lat': 'latitude',
        'lon': 'longitude',
    }
}
"""Excel variable names"""

LOGGER = logging.getLogger(__name__)


class Centroids():
    """Contains raster or vector centroids.

    Attributes
    ----------
    meta : dict, optional
        rasterio meta dictionary containing raster properties: width, height, crs and transform
        must be present at least. The affine ransformation needs to be shearless (only stretching)
        and have positive x- and negative y-orientation.
    lat : np.array, optional
        latitude of size size
    lon : np.array, optional
        longitude of size size
    geometry : gpd.GeoSeries, optional
        contains lat and lon crs. Might contain geometry points for lat and lon
    area_pixel : np.array, optional
        area of size size
    dist_coast : np.array, optional
        distance to coast of size size
    on_land : np.array, optional
        on land (True) and on sea (False) of size size
    region_id : np.array, optional
        country region code of size size
    elevation : np.array, optional
        elevation of size size
    """

    vars_check = {'lat', 'lon', 'geometry', 'area_pixel', 'dist_coast',
                  'on_land', 'region_id', 'elevation'}
    """Variables whose size will be checked"""

    def __init__(self):
        """Initialize to None raster and vector. Default crs=DEF_CRS"""
        self.meta = dict()
        self.geometry = gpd.GeoSeries(crs=DEF_CRS)
        self.lat = np.array([])
        self.lon = np.array([])
        self.area_pixel = np.array([])
        self.dist_coast = np.array([])
        self.on_land = np.array([])
        self.region_id = np.array([])
        self.elevation = np.array([])

    def check(self):
        """Check integrity of stored information.

        Checks that either `meta` attribute is set, or `lat`, `lon` and `geometry.crs`.
        Checks sizes of (optional) data attributes."""
        n_centr = self.size
        for var_name, var_val in self.__dict__.items():
            if var_name in self.vars_check:
                if var_val.size > 0 and var_val.size != n_centr:
                    raise ValueError(f'Wrong {var_name} size: {n_centr} != {var_val.size}.')
        if self.meta:
            for name in ['width', 'height', 'crs', 'transform']:
                if name not in self.meta.keys():
                    raise ValueError('Missing meta information: %s' % name)
            xres, xshear, _xoff, yshear, yres, _yoff = self.meta['transform'][:6]
            if xshear != 0 or yshear != 0:
                raise ValueError('Affine transformations with shearing components are not '
                                 'supported.')
            if yres > 0 or xres < 0:
                raise ValueError('Affine transformations with positive y-orientation '
                                 'or negative x-orientation are not supported.')

    def equal(self, centr):
        """Return True if two centroids equal, False otherwise

        Parameters
        ----------
        centr : Centroids
            centroids to compare

        Returns
        -------
        eq : bool
        """
        if self.meta and centr.meta:
            return (u_coord.equal_crs(self.meta['crs'], centr.meta['crs'])
                    and self.meta['height'] == centr.meta['height']
                    and self.meta['width'] == centr.meta['width']
                    and self.meta['transform'] == centr.meta['transform'])
        return (u_coord.equal_crs(self.crs, centr.crs)
                and self.lat.shape == centr.lat.shape
                and self.lon.shape == centr.lon.shape
                and np.allclose(self.lat, centr.lat)
                and np.allclose(self.lon, centr.lon))

    @staticmethod
    def from_base_grid(land=False, res_as=360, base_file=None):
        """Initialize from base grid data provided with CLIMADA

        Parameters
        ----------
        land : bool, optional
            If True, restrict to grid points on land. Default: False.
        res_as : int, optional
            Base grid resolution in arc-seconds (one of 150, 360). Default: 360.
        base_file : str, optional
            If set, read this file instead of one provided with climada.
        """
        if base_file is None:
            base_file = NATEARTH_CENTROIDS[res_as]

        centroids = Centroids.from_hdf5(base_file)
        if centroids.meta:
            xres, xshear, xoff, yshear, yres, yoff = centroids.meta['transform'][:6]
            shape = (centroids.meta['height'], centroids.meta['width'])
            if yres > 0:
                # make sure y-orientation is negative
                centroids.meta['transform'] = rasterio.Affine(xres, xshear, xoff, yshear,
                                                              -yres, yoff + (shape[0] - 1) * yres)
                # flip y-axis in data arrays
                for name in ["region_id", "dist_coast"]:
                    if not hasattr(centroids, name):
                        continue
                    data = getattr(centroids, name)
                    if data.size == 0:
                        continue
                    setattr(centroids, name, np.flipud(data.reshape(shape)).reshape(-1))
        if land:
            land_reg_ids = list(range(1, 1000))
            land_reg_ids.remove(10)  # Antarctica
            centroids = centroids.select(reg_id=land_reg_ids)

        centroids.check()
        return centroids

    @classmethod
    def from_geodataframe(cls, gdf, geometry_alias='geom'):
        """Create Centroids instance from GeoDataFrame.

        The geometry, lat, and lon attributes are set from the GeoDataFrame.geometry attribute,
        while the columns are copied as attributes to the Centroids object in the form of
        numpy.ndarrays using pandas.Series.to_numpy. The Series dtype will thus be respected.

        Columns named lat or lon are ignored, as they would overwrite the coordinates extracted
        from the point features. If the geometry attribute bears an alias, it can be dropped by
        setting the geometry_alias parameter.

        If the GDF includes a region_id column, but no on_land column, then on_land=True is
        inferred for those centroids that have a set region_id.

        Example
        -------
        >>> gdf = geopandas.read_file('centroids.shp')
        >>> gdf.region_id = gdf.region_id.astype(int)  # type coercion
        >>> centroids = Centroids.from_geodataframe(gdf)

        Parameters
        ----------
        gdf : GeoDataFrame
            Where the geometry column needs to consist of point features. See above for details on
            processing.
        geometry_alias : str, opt
            Alternate name for the geometry column; dropped to avoid duplicate assignment.

        Returns
        -------
        centr : Centroids
            Centroids with data from given GeoDataFrame
        """
        centroids = cls()

        centroids.geometry = gdf.geometry
        centroids.lat = gdf.geometry.y.to_numpy(copy=True)
        centroids.lon = gdf.geometry.x.to_numpy(copy=True)

        for col in gdf.columns:
            if col in [geometry_alias, 'geometry', 'lat', 'lon']:
                continue  # skip these, because they're already set above
            val = gdf[col].to_numpy(copy=True)
            setattr(centroids, col, val)

        if centroids.on_land.size == 0:
            try:
                centroids.on_land = ~np.isnan(centroids.region_id)
            except KeyError:
                pass

        return centroids

    def set_raster_from_pix_bounds(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_pix_bounds instead."""
        LOGGER.warning("The use of Centroids.set_raster_from_pix_bounds is deprecated. "
                       "Use Centroids.from_pix_bounds instead.")
        self.__dict__ = Centroids.from_pix_bounds(*args, **kwargs).__dict__

    @classmethod
    def from_pix_bounds(cls, xf_lat, xo_lon, d_lat, d_lon, n_lat, n_lon, crs=DEF_CRS):
        """Create Centroids object with meta attribute according to pixel border data.

        Parameters
        ----------
        xf_lat : float
            upper latitude (top)
        xo_lon : float
            left longitude
        d_lat : float
            latitude step (negative)
        d_lon : float
            longitude step (positive)
        n_lat : int
            number of latitude points
        n_lon : int
            number of longitude points
        crs : dict() or rasterio.crs.CRS, optional
            CRS. Default: DEF_CRS

        Returns
        -------
        centr : Centroids
            Centroids with meta according to given pixel border data.
        """
        centr = cls()
        centr.meta = {
            'dtype': 'float32',
            'width': n_lon,
            'height': n_lat,
            'crs': crs,
            'transform': rasterio.Affine(d_lon, 0.0, xo_lon, 0.0, d_lat, xf_lat),
        }
        return centr

    def set_raster_from_pnt_bounds(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_pnt_bounds instead."""
        LOGGER.warning("The use of Centroids.set_raster_from_pnt_bounds is deprecated. "
                       "Use Centroids.from_pnt_bounds instead.")
        self.__dict__ = Centroids.from_pnt_bounds(*args, **kwargs).__dict__

    @classmethod
    def from_pnt_bounds(cls, points_bounds, res, crs=DEF_CRS):
        """Create Centroids object with meta attribute according to points border data.

        raster border = point border + res/2

        Parameters
        ----------
        points_bounds : tuple
            points' lon_min, lat_min, lon_max, lat_max
        res : float
            desired resolution in same units as points_bounds
        crs : dict() or rasterio.crs.CRS, optional
            CRS. Default: DEF_CRS

        Returns
        -------
        centr : Centroids
            Centroids with meta according to given points border data.
        """
        centr = cls()
        rows, cols, ras_trans = u_coord.pts_to_raster_meta(points_bounds, (res, -res))
        centr.meta = {
            'width': cols,
            'height': rows,
            'crs': crs,
            'transform': ras_trans,
        }
        return centr

    def set_lat_lon(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_lat_lon instead."""
        LOGGER.warning("The use of Centroids.set_lat_lon is deprecated. "
                       "Use Centroids.from_lat_lon instead.")
        self.__dict__ = Centroids.from_lat_lon(*args, **kwargs).__dict__

    @classmethod
    def from_lat_lon(cls, lat, lon, crs=DEF_CRS):
        """Create Centroids object from given latitude, longitude and CRS.

        Parameters
        ----------
        lat : np.array
            latitude
        lon : np.array
            longitude
        crs : dict() or rasterio.crs.CRS, optional
            CRS. Default: DEF_CRS

        Returns
        -------
        centr : Centroids
            Centroids with points according to given coordinates
        """
        centr = cls()
        centr.lat = np.asarray(lat)
        centr.lon = np.asarray(lon)
        centr.geometry = gpd.GeoSeries(crs=crs)
        return centr

    def set_raster_file(self, file_name, band=None, **kwargs):
        """This function is deprecated, use Centroids.from_raster_file
        and Centroids.values_from_raster_files instead."""
        LOGGER.warning("The use of Centroids.set_raster_file is deprecated. "
                       "Use Centroids.from_raster_file and "
                       "Centroids.values_from_raster_files instead.")
        if not self.meta:
            self.__dict__ = Centroids.from_raster_file(file_name, **kwargs).__dict__
        return self.values_from_raster_files([file_name], band=band, **kwargs)

    @classmethod
    def from_raster_file(cls, file_name, src_crs=None, window=False,
                         geometry=False, dst_crs=False, transform=None, width=None,
                         height=None, resampling=Resampling.nearest):
        """Create a new Centroids object from a raster file

        Select region using window or geometry. Reproject input by providing
        dst_crs and/or (transform, width, height).

        Parameters
        ----------
        file_name : str
            path of the file
        src_crs : crs, optional
            source CRS. Provide it if error without it.
        window : rasterio.windows.Window, optional
            window to read
        geometry : shapely.geometry, optional
            consider pixels only in shape
        dst_crs : crs, optional
            reproject to given crs
        transform : rasterio.Affine
            affine transformation to apply
        wdith : float
            number of lons for transform
        height : float
            number of lats for transform
        resampling : rasterio.warp,.Resampling optional
            resampling function used for reprojection to dst_crs

        Returns
        -------
        centr : Centroids
            Centroids with meta attribute according to the given raster file
        """
        centr = cls()
        centr.meta, _ = u_coord.read_raster(
            file_name, [1], src_crs, window, geometry, dst_crs,
            transform, width, height, resampling)
        return centr

    def values_from_raster_files(self, file_names, band=None, src_crs=None, window=False,
                                 geometry=False, dst_crs=False, transform=None, width=None,
                                 height=None, resampling=Resampling.nearest):
        """Read raster of bands and set 0 values to the masked ones.

        Each band is an event. Select region using window or geometry. Reproject input by proving
        dst_crs and/or (transform, width, height).

        Parameters
        ----------
        file_names : str
            path of the file
        band : list(int), optional
            band number to read. Default: [1]
        src_crs : crs, optional
            source CRS. Provide it if error without it.
        window : rasterio.windows.Window, optional
            window to read
        geometry : shapely.geometry, optional
            consider pixels only in shape
        dst_crs : crs, optional
            reproject to given crs
        transform : rasterio.Affine
            affine transformation to apply
        wdith : float
            number of lons for transform
        height : float
            number of lats for transform
        resampling : rasterio.warp,.Resampling optional
            resampling function used for reprojection to dst_crs

        Raises
        ------
        ValueError

        Returns
        -------
        inten : scipy.sparse.csr_matrix
            Each row is an event.
        """
        if band is None:
            band = [1]

        values = []
        for file_name in file_names:
            tmp_meta, data = u_coord.read_raster(
                file_name, band, src_crs, window, geometry, dst_crs,
                transform, width, height, resampling)
            if (tmp_meta['crs'] != self.meta['crs']
                    or tmp_meta['transform'] != self.meta['transform']
                    or tmp_meta['height'] != self.meta['height']
                    or tmp_meta['width'] != self.meta['width']):
                raise ValueError('Raster data is inconsistent with contained raster.')
            values.append(sparse.csr_matrix(data))

        return sparse.vstack(values, format='csr')


    def set_vector_file(self, file_name, inten_name=None, **kwargs):
        """This function is deprecated, use Centroids.from_vector_file
        and Centroids.values_from_vector_files instead."""
        LOGGER.warning("The use of Centroids.set_vector_file is deprecated. "
                       "Use Centroids.from_vector_file and "
                       "Centroids.values_from_vector_files instead.")
        if not self.geometry.any():
            self.__dict__ = Centroids.from_vector_file(file_name, **kwargs).__dict__
        return self.values_from_vector_files([file_name], val_names=inten_name, **kwargs)

    @classmethod
    def from_vector_file(cls, file_name, dst_crs=None):
        """Create Centroids object from vector file (any format supported by fiona).

        Parameters
        ----------
        file_name : str
            vector file with format supported by fiona and 'geometry' field.
        dst_crs : crs, optional
            reproject to given crs

        Returns
        -------
        centr : Centroids
            Centroids with points according to the given vector file
        """
        centr = cls()
        centr.lat, centr.lon, centr.geometry, _ = u_coord.read_vector(
            file_name, [], dst_crs=dst_crs)
        return centr

    def values_from_vector_files(self, file_names, val_names=None, dst_crs=None):
        """Read intensity or other data from vector files, making sure that geometry is compatible.

        If the geometry of the shapes in any of the given files does not agree with the
        geometry of this Centroids instance, a ValueError is raised.

        Parameters
        ----------
        file_names : list(str)
            vector files with format supported by fiona and 'geometry' field.
        val_names : list(str), optional
            list of names of the columns of the values. Default: ['intensity']
        dst_crs : crs, optional
            reproject to given crs

        Raises
        ------
        ValueError

        Returns
        -------
        values : scipy.sparse.csr_matrix
            Sparse array of shape (len(val_name), len(geometry)).
        """
        if val_names is None:
            val_names = ['intensity']

        values = []
        for file_name in file_names:
            tmp_lat, tmp_lon, tmp_geometry, data = u_coord.read_vector(
                file_name, val_names, dst_crs=dst_crs)
            if not (u_coord.equal_crs(tmp_geometry.crs, self.geometry.crs)
                    and np.allclose(tmp_lat, self.lat)
                    and np.allclose(tmp_lon, self.lon)):
                raise ValueError('Vector data inconsistent with contained vector.')
            values.append(sparse.csr_matrix(data))

        return sparse.vstack(values, format='csr')

    def read_mat(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_mat instead."""
        LOGGER.warning("The use of Centroids.read_mat is deprecated."
                       "Use Centroids.from_mat instead.")
        self.__dict__ = Centroids.from_mat(*args, **kwargs).__dict__

    @classmethod
    def from_mat(cls, file_name, var_names=None):
        """Read centroids from CLIMADA's MATLAB version.

        Parameters
        ----------
        file_name : str
            absolute or relative file name
        var_names : dict, optional
            name of the variables

        Raises
        ------
        KeyError

        Returns
        -------
        centr : Centroids
            Centroids with data from the given file
        """
        LOGGER.info('Reading %s', file_name)
        if var_names is None:
            var_names = DEF_VAR_MAT

        cent = u_hdf5.read(file_name)
        # Try open encapsulating variable FIELD_NAMES
        num_try = 0
        for field in var_names['field_names']:
            try:
                cent = cent[field]
                break
            except KeyError:
                num_try += 1
        if num_try == len(var_names['field_names']):
            LOGGER.warning("Variables are not under: %s.", var_names['field_names'])

        try:
            cen_lat = np.squeeze(cent[var_names['var_name']['lat']])
            cen_lon = np.squeeze(cent[var_names['var_name']['lon']])
            centr = cls.from_lat_lon(cen_lat, cen_lon)

            try:
                centr.dist_coast = np.squeeze(cent[var_names['var_name']['dist_coast']])
            except KeyError:
                pass
            try:
                centr.region_id = np.squeeze(cent[var_names['var_name']['region_id']])
            except KeyError:
                pass
        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err

        return centr

    def read_excel(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_excel instead."""
        LOGGER.warning("The use of Centroids.read_excel is deprecated."
                       "Use Centroids.from_excel instead.")
        self.__dict__ = Centroids.from_excel(*args, **kwargs).__dict__

    @classmethod
    def from_excel(cls, file_name, var_names=None):
        """Generate a new centroids object from an excel file with column names in var_names.

        Parameters
        ----------
        file_name : str
            absolute or relative file name
        var_names : dict, default
            name of the variables

        Raises
        ------
        KeyError

        Returns
        -------
        centr : Centroids
            Centroids with data from the given file
        """
        LOGGER.info('Reading %s', file_name)
        if var_names is None:
            var_names = DEF_VAR_EXCEL

        try:
            dfr = pd.read_excel(file_name, var_names['sheet_name'])
            centr = cls.from_lat_lon(dfr[var_names['col_name']['lat']],
                                     dfr[var_names['col_name']['lon']])
            try:
                centr.region_id = dfr[var_names['col_name']['region_id']]
            except KeyError:
                pass

        except KeyError as err:
            raise KeyError("Not existing variable: %s" % str(err)) from err

        return centr

    def clear(self):
        """Clear vector and raster data."""
        self.__init__()

    def append(self, centr):
        """Append centroids points.

        If centr or self are rasters they are converted to points first using
        Centroids.set_meta_to_lat_lon. Note that self is modified in-place,
        and meta is set to {}. Thus, raster information in self is lost.

        Note: this is a wrapper for centroids.union.

        Parameters
        ----------
        centr : Centroids
            Centroids to append. The centroids need to have the same CRS.

        See Also
        --------
        union : Union of Centroid objects.
        """
        self.__dict__.update(self.union(centr).__dict__)


    def union(self, *others):
        """
        Create the union of centroids from the inputs.

        The centroids are combined together point by point.
        Rasters are converted to points and raster information is lost
        in the output. All centroids must have the same CRS.

        In any case, the attribute .geometry is computed for all centroids.
        This requires a CRS to be defined. If Centroids.crs is None, the
        default DEF_CRS is set for all centroids (self and others).

        When at least one centroids has one of the following property
        defined, it is also computed for all others.
        .area_pixel, .dist_coast, .on_land, .region_id, .elevetaion'

        !Caution!: the input objects (self and others) are modified in place.
        Missing properties are added, existing ones are not overwritten.

        Parameters
        ----------
        others : any number of climada.hazard.Centroids()
            Centroids to form the union with

        Returns
        -------
        centroids : Centroids
            Centroids containing the union of the centroids in others.

        Raises
        ------
        ValueError
        """
        # restrict to non-empty centroids
        cent_list = [c for c in (self,) + others if c.size > 0 or c.meta]
        if len(cent_list) == 0 or len(others) == 0:
            return copy.deepcopy(self)

        # check if all centroids are identical
        if all([cent_list[0].equal(cent) for cent in cent_list[1:]]):
            return copy.deepcopy(cent_list[0])

        # convert all raster centroids to point centroids
        for cent in cent_list:
            if cent.meta and not cent.lat.any():
                cent.set_meta_to_lat_lon()

        # make sure that all Centroids have the same CRS
        for cent in cent_list:
            if cent.crs is None:
                cent.geometry = cent.geometry.set_crs(DEF_CRS)
            if not u_coord.equal_crs(cent.crs, cent_list[0].crs):
                raise ValueError('In a union, all Centroids need to have the same CRS: '
                                 f'{cent.crs} != {cent_list[0].crs}')

        # set attributes that are missing in some but defined in others
        for attr in ["geometry", "area_pixel", "dist_coast", "on_land", "region_id", "elevation"]:
            if np.any([getattr(cent, attr).size > 0 for cent in cent_list]):
                for cent in cent_list:
                    if not getattr(cent, attr).size > 0:
                        fun_name = f"set_{attr}{'_points' if attr == 'geometry' else ''}"
                        getattr(Centroids, fun_name)(cent)

        # create new Centroids object and set concatenated attributes
        centroids = Centroids()
        centroids.meta = {}
        for attr_name, attr_val in vars(cent_list[0]).items():
            if isinstance(attr_val, np.ndarray) and attr_val.ndim == 1:
                attr_val_list = [getattr(cent, attr_name) for cent in cent_list]
                setattr(centroids, attr_name, np.hstack(attr_val_list))
            elif isinstance(attr_val, gpd.GeoSeries):
                attr_val_list = [getattr(cent, attr_name) for cent in cent_list]
                setattr(centroids, attr_name, pd.concat(attr_val_list, ignore_index=True))

        # finally, remove duplicate points
        return centroids.remove_duplicate_points()

    def get_closest_point(self, x_lon, y_lat, scheduler=None):
        """Returns closest centroid and its index to a given point.

        Parameters
        ----------
        x_lon : float
            x coord (lon)
        y_lat : float
            y coord (lat)
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”

        Returns
        -------
        x_close : float
            x-coordinate (longitude) of closest centroid.
        y_close : float
            y-coordinate (latitude) of closest centroids.
        idx_close : int
            Index of centroid in internal ordering of centroids.
        """
        if self.meta:
            if not self.lat.size or not self.lon.size:
                self.set_meta_to_lat_lon()
            i_lat, i_lon = rasterio.transform.rowcol(self.meta['transform'], x_lon, y_lat)
            i_lat = np.clip(i_lat, 0, self.meta['height'] - 1)
            i_lon = np.clip(i_lon, 0, self.meta['width'] - 1)
            close_idx = int(i_lat * self.meta['width'] + i_lon)
        else:
            self.set_geometry_points(scheduler)
            close_idx = self.geometry.distance(Point(x_lon, y_lat)).values.argmin()
        return self.lon[close_idx], self.lat[close_idx], close_idx

    def set_region_id(self, scheduler=None):
        """Set region_id as country ISO numeric code attribute for every pixel or point.

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”
        """
        ne_geom = self._ne_crs_geom(scheduler)
        LOGGER.debug('Setting region_id %s points.', str(self.lat.size))
        self.region_id = u_coord.get_country_code(
            ne_geom.geometry[:].y.values, ne_geom.geometry[:].x.values)

    def set_area_pixel(self, min_resol=1.0e-8, scheduler=None):
        """Set `area_pixel` attribute for every pixel or point (area in m*m).

        Parameters
        ----------
        min_resol : float, optional
            if centroids are points, use this minimum resolution in lat and lon. Default: 1.0e-8
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”
        """
        if self.meta:
            if hasattr(self.meta['crs'], 'linear_units') and \
            str.lower(self.meta['crs'].linear_units) in ['m', 'metre', 'meter']:
                self.area_pixel = np.zeros((self.meta['height'], self.meta['width']))
                self.area_pixel *= abs(self.meta['transform'].a) * abs(self.meta['transform'].e)
                return
            if abs(abs(self.meta['transform'].a) -
                   abs(self.meta['transform'].e)) > 1.0e-5:
                raise ValueError('Area can not be computed for not squared pixels.')
            res = self.meta['transform'].a
        else:
            res = u_coord.get_resolution(self.lat, self.lon, min_resol=min_resol)
            res = np.abs(res).min()
        self.set_geometry_points(scheduler)
        LOGGER.debug('Setting area_pixel %s points.', str(self.lat.size))
        xy_pixels = self.geometry.buffer(res / 2).envelope
        if PROJ_CEA == self.geometry.crs:
            self.area_pixel = xy_pixels.area.values
        else:
            self.area_pixel = xy_pixels.to_crs(crs={'proj': 'cea'}).area.values

    def set_area_approx(self, min_resol=1.0e-8):
        """Set `area_pixel` attribute for every pixel or point (approximate area in m*m).

        Values are differentiated per latitude. Faster than `set_area_pixel`.

        Parameters
        ----------
        min_resol : float, optional
            if centroids are points, use this minimum resolution in lat and lon. Default: 1.0e-8
        """
        if self.meta:
            if hasattr(self.meta['crs'], 'linear_units') and \
            str.lower(self.meta['crs'].linear_units) in ['m', 'metre', 'meter']:
                self.area_pixel = np.zeros((self.meta['height'], self.meta['width']))
                self.area_pixel *= abs(self.meta['transform'].a) * abs(self.meta['transform'].e)
                return
            res_lat, res_lon = self.meta['transform'].e, self.meta['transform'].a
            lat_unique = np.arange(self.meta['transform'].f + res_lat / 2,
                                   self.meta['transform'].f + self.meta['height'] * res_lat,
                                   res_lat)
            lon_unique_len = self.meta['width']
            res_lat = abs(res_lat)
        else:
            res_lat, res_lon = np.abs(
                u_coord.get_resolution(self.lat, self.lon, min_resol=min_resol))
            lat_unique = np.array(np.unique(self.lat))
            lon_unique_len = len(np.unique(self.lon))
            if PROJ_CEA == self.geometry.crs:
                self.area_pixel = np.repeat(res_lat * res_lon, lon_unique_len)
                return

        LOGGER.debug('Setting area_pixel approx %s points.', str(self.lat.size))
        res_lat = res_lat * ONE_LAT_KM * 1000
        res_lon = res_lon * ONE_LAT_KM * 1000 * np.cos(np.radians(lat_unique))
        area_approx = np.repeat(res_lat * res_lon, lon_unique_len)
        if area_approx.size == self.size:
            self.area_pixel = area_approx
        else:
            raise ValueError('Pixel area of points can not be computed.')

    def set_elevation(self, topo_path):
        """Set elevation attribute for every pixel or point in meters.

        Parameters
        ----------
        topo_path : str
            Path to a raster file containing gridded elevation data.
        """
        if not self.coord.size:
            self.set_meta_to_lat_lon()
        self.elevation = u_coord.read_raster_sample(topo_path, self.lat, self.lon)

    def set_dist_coast(self, signed=False, precomputed=False, scheduler=None):
        """Set dist_coast attribute for every pixel or point in meters.

        Parameters
        ----------
        signed : bool
            If True, use signed distances (positive off shore and negative on land). Default: False.
        precomputed : bool
            If True, use precomputed distances (from NASA). Default: False.
        scheduler : str
            Used for dask map_partitions. "threads", "synchronous" or "processes"
        """
        if (not self.lat.size or not self.lon.size) and not self.meta:
            LOGGER.warning('No lat/lon, no meta, nothing to do!')
            return
        if precomputed:
            if not self.lat.size or not self.lon.size:
                self.set_meta_to_lat_lon()
            self.dist_coast = u_coord.dist_to_coast_nasa(
                self.lat, self.lon, highres=True, signed=signed)
        else:
            ne_geom = self._ne_crs_geom(scheduler)
            LOGGER.debug('Computing distance to coast for %s centroids.', str(self.lat.size))
            self.dist_coast = u_coord.dist_to_coast(ne_geom, signed=signed)

    def set_on_land(self, scheduler=None):
        """Set on_land attribute for every pixel or point.

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”
        """
        ne_geom = self._ne_crs_geom(scheduler)
        LOGGER.debug('Setting on_land %s points.', str(self.lat.size))
        self.on_land = u_coord.coord_on_land(
            ne_geom.geometry[:].y.values, ne_geom.geometry[:].x.values)

    def remove_duplicate_points(self, scheduler=None):
        """Return Centroids with removed duplicated points

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”

        Returns
        -------
        cen : Centroids
            Sub-selection of this object.
        """
        if not self.lat.any() and not self.meta:
            return self
        if self.lat.size > 0:
            coords_view = self.coord.astype(np.float64).view(dtype='float64,float64')
            sel_cen = np.sort(np.unique(coords_view, return_index=True)[1])
        else:
            geom_wkb = self.geometry.apply(lambda geom: geom.wkb)
            sel_cen = geom_wkb.drop_duplicates().index
        return self.select(sel_cen=sel_cen)

    def select(self, reg_id=None, extent=None, sel_cen=None):
        """Return Centroids with points in the given reg_id or within mask

        Parameters
        ----------
        reg_id : int
            region to filter according to region_id values
        extent : tuple
            Format (min_lon, max_lon, min_lat, max_lat) tuple.
            If min_lon > lon_max, the extend crosses the antimeridian and is
            [lon_max, 180] + [-180, lon_min]
            Borders are inclusive.
        sel_cen : np.array
            1-dim mask, overrides reg_id and extent

        Returns
        -------
        cen : Centroids
            Sub-selection of this object
        """
        if sel_cen is None:
            sel_cen = self.select_mask(reg_id=reg_id, extent=extent)

        if not self.lat.size or not self.lon.size:
            self.set_meta_to_lat_lon()

        centr = Centroids.from_lat_lon(self.lat[sel_cen], self.lon[sel_cen], self.geometry.crs)
        if self.area_pixel.size:
            centr.area_pixel = self.area_pixel[sel_cen]
        if self.region_id.size:
            centr.region_id = self.region_id[sel_cen]
        if self.on_land.size:
            centr.on_land = self.on_land[sel_cen]
        if self.dist_coast.size:
            centr.dist_coast = self.dist_coast[sel_cen]
        return centr

    def select_mask(self, reg_id=None, extent=None):
        """
        Make mask of selected centroids

        Parameters
        ----------
        reg_id : int
            region to filter according to region_id values
        extent : tuple
            Format (min_lon, max_lon, min_lat, max_lat) tuple.
            If min_lon > lon_max, the extend crosses the antimeridian and is
            [lon_max, 180] + [-180, lon_min]
            Borders are inclusive.

        Returns
        -------
        sel_cen : 1d array of booleans
            1d mask of selected centroids

        """
        sel_cen = np.ones(self.size, dtype=bool)
        if reg_id is not None:
            sel_cen &= np.isin(self.region_id, reg_id)
        if extent is not None:
            lon_min, lon_max, lat_min, lat_max = extent
            lon_max += 360 if lon_min > lon_max else 0
            lon_normalized = u_coord.lon_normalize(
                self.lon.copy(), center=0.5 * (lon_min + lon_max))
            sel_cen &= (
              (lon_normalized >= lon_min) & (lon_normalized <= lon_max) &
              (self.lat >= lat_min) & (self.lat <= lat_max)
            )
        return sel_cen

    def set_lat_lon_to_meta(self, min_resol=1.0e-8):
        """Compute meta from lat and lon values.

        Parameters
        ----------
        min_resol : float, optional
            Minimum centroids resolution to use in the raster. Default: 1.0e-8.
        """
        res = u_coord.get_resolution(self.lon, self.lat, min_resol=min_resol)
        rows, cols, ras_trans = u_coord.pts_to_raster_meta(self.total_bounds, res)
        LOGGER.debug('Resolution points: %s', str(res))
        self.meta = {
            'width': cols,
            'height': rows,
            'crs': self.crs,
            'transform': ras_trans,
        }

    def set_meta_to_lat_lon(self):
        """Compute lat and lon of every pixel center from meta raster."""
        if self.meta:
            xgrid, ygrid = u_coord.raster_to_meshgrid(
                self.meta['transform'], self.meta['width'], self.meta['height'])
            self.lon = xgrid.flatten()
            self.lat = ygrid.flatten()
            self.geometry = gpd.GeoSeries(crs=self.meta['crs'])

    def plot(self, axis=None, figsize=(9, 13), **kwargs):
        """Plot centroids scatter points over earth.

        Parameters
        ----------
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        figsize: (float, float), optional
            figure size for plt.subplots
            The default is (9, 13)
        kwargs : optional
            arguments for scatter matplotlib function

        Returns
        -------
        axis : matplotlib.axes._subplots.AxesSubplot
        """
        if self.meta and not self.coord.size:
            self.set_meta_to_lat_lon()
        pad = np.abs(u_coord.get_resolution(self.lat, self.lon)).min()

        proj_data, _ = u_plot.get_transformation(self.crs)
        proj_plot = proj_data
        if isinstance(proj_data, ccrs.PlateCarree):
            # use different projections for plot and data to shift the central lon in the plot
            xmin, ymin, xmax, ymax = u_coord.latlon_bounds(self.lat, self.lon, buffer=pad)
            proj_plot = ccrs.PlateCarree(central_longitude=0.5 * (xmin + xmax))
        else:
            xmin, ymin, xmax, ymax = (self.lon.min() - pad, self.lat.min() - pad,
                                      self.lon.max() + pad, self.lat.max() + pad)

        if not axis:
            _, axis, _fontsize = u_plot.make_map(proj=proj_plot, figsize=figsize)

        axis.set_extent((xmin, xmax, ymin, ymax), crs=proj_data)
        u_plot.add_shapes(axis)
        axis.scatter(self.lon, self.lat, transform=proj_data, **kwargs)
        plt.tight_layout()
        return axis

    def calc_pixels_polygons(self, scheduler=None):
        """Return a gpd.GeoSeries with a polygon for every pixel

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”

        Returns
        -------
        geo : gpd.GeoSeries
        """
        if not self.meta:
            self.set_lat_lon_to_meta()
        if abs(abs(self.meta['transform'].a) -
               abs(self.meta['transform'].e)) > 1.0e-5:
            raise ValueError('Area can not be computed for not squared pixels.')
        self.set_geometry_points(scheduler)
        return self.geometry.buffer(self.meta['transform'].a / 2).envelope

    def empty_geometry_points(self):
        """Removes all points in geometry.

        Useful when centroids is used in multiprocessing function."""
        self.geometry = gpd.GeoSeries(crs=self.geometry.crs)

    def write_hdf5(self, file_data):
        """Write centroids attributes into hdf5 format.

        Parameters
        ----------
        file_data : str or h5
            If string, path to write data. If h5 object, the datasets will be generated there.
        """
        if isinstance(file_data, str):
            LOGGER.info('Writting %s', file_data)
            data = h5py.File(file_data, 'w')
        else:
            data = file_data
        str_dt = h5py.special_dtype(vlen=str)
        for centr_name, centr_val in self.__dict__.items():
            if isinstance(centr_val, np.ndarray):
                data.create_dataset(centr_name, data=centr_val, compression="gzip")
            if centr_name == 'meta' and centr_val:
                centr_meta = data.create_group(centr_name)
                for key, value in centr_val.items():
                    if key not in ('crs', 'transform'):
                        if not isinstance(value, str):
                            centr_meta.create_dataset(key, (1,), data=value, dtype=type(value))
                        else:
                            hf_str = centr_meta.create_dataset(key, (1,), dtype=str_dt)
                            hf_str[0] = value
                    elif key == 'transform':
                        centr_meta.create_dataset(
                            key, (6,),
                            data=[value.a, value.b, value.c, value.d, value.e, value.f],
                            dtype=float)
        hf_str = data.create_dataset('crs', (1,), dtype=str_dt)
        hf_str[0] = CRS.from_user_input(self.crs).to_wkt()

        if isinstance(file_data, str):
            data.close()

    def read_hdf5(self, *args, **kwargs):
        """This function is deprecated, use Centroids.from_hdf5 instead."""
        LOGGER.warning("The use of Centroids.read_hdf5 is deprecated."
                       "Use Centroids.from_hdf5 instead.")
        self.__dict__ = Centroids.from_hdf5(*args, **kwargs).__dict__

    @classmethod
    def from_hdf5(cls, file_data):
        """Create a centroids object from a HDF5 file.

        Parameters
        ----------
        file_data : str or h5
            If string, path to read data. If h5 object, the datasets will be read from there.

        Returns
        -------
        centr : Centroids
            Centroids with data from the given file
        """
        if isinstance(file_data, (str, Path)):
            LOGGER.info('Reading %s', file_data)
            data = h5py.File(file_data, 'r')
        else:
            data = file_data
        centr = None
        crs = DEF_CRS
        if data.get('crs'):
            crs = u_coord.to_crs_user_input(data.get('crs')[0])
        if data.get('lat') and data.get('lat').size:
            centr = cls.from_lat_lon(
                np.array(data.get('lat')),
                np.array(data.get('lon')),
                crs=crs)
        elif data.get('latitude') and data.get('latitude').size:
            centr = cls.from_lat_lon(
                np.array(data.get('latitude')),
                np.array(data.get('longitude')),
                crs=crs)
        else:
            centr_meta = data.get('meta')
            centr = cls()
            centr.meta['crs'] = crs
            for key, value in centr_meta.items():
                if key != 'transform':
                    centr.meta[key] = value[0]
                else:
                    centr.meta[key] = rasterio.Affine(*value)
        for centr_name in data.keys():
            if centr_name not in ('crs', 'lat', 'lon', 'meta'):
                setattr(centr, centr_name, np.array(data.get(centr_name)))
        if isinstance(file_data, str):
            data.close()
        return centr

    @property
    def crs(self):
        """Get CRS of raster or vector."""
        if self.meta:
            return self.meta['crs']
        elif self.geometry.crs:
            return self.geometry.crs
        return DEF_CRS

    @property
    def size(self):
        """Get number of pixels or points."""
        if self.meta:
            return int(self.meta['height'] * self.meta['width'])
        return self.lat.size

    @property
    def shape(self):
        """Get shape of rastered data."""
        try:
            if self.meta:
                return (self.meta['height'], self.meta['width'])
            return (np.unique(self.lat).size, np.unique(self.lon).size)
        except AttributeError:
            return ()

    @property
    def total_bounds(self):
        """Get total bounds (left, bottom, right, top)."""
        if self.meta:
            left = self.meta['transform'].xoff
            right = left + self.meta['transform'][0] * self.meta['width']
            if left > right:
                left, right = right, left
            top = self.meta['transform'].yoff
            bottom = top + self.meta['transform'][4] * self.meta['height']
            if bottom > top:
                bottom, top = top, bottom
            return left, bottom, right, top
        return self.lon.min(), self.lat.min(), self.lon.max(), self.lat.max()

    @property
    def coord(self):
        """Get [lat, lon] array."""
        return np.stack([self.lat, self.lon], axis=1)

    def set_geometry_points(self, scheduler=None):
        """Set `geometry` attribute with Points from `lat`/`lon` attributes.

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”
        """
        def apply_point(df_exp):
            return df_exp.apply((lambda row: Point(row.longitude, row.latitude)), axis=1)
        if not self.geometry.size:
            LOGGER.info('Convert centroids to GeoSeries of Point shapes.')
            if (not self.lat.any() or not self.lon.any()) and self.meta:
                self.set_meta_to_lat_lon()
            if not scheduler:
                self.geometry = gpd.GeoSeries(
                    gpd.points_from_xy(self.lon, self.lat), crs=self.geometry.crs)
            else:
                import dask.dataframe as dd
                from multiprocessing import cpu_count
                ddata = dd.from_pandas(self, npartitions=cpu_count())
                self.geometry = (ddata
                                 .map_partitions(apply_point, meta=Point)
                                 .compute(scheduler=scheduler))

    def _ne_crs_geom(self, scheduler=None):
        """Return `geometry` attribute in the CRS of Natural Earth.

        Parameters
        ----------
        scheduler : str
            used for dask map_partitions. “threads”, “synchronous” or “processes”

        Returns
        -------
        geo : gpd.GeoSeries
        """
        if not self.lat.size or not self.lon.size:
            self.set_meta_to_lat_lon()
        if u_coord.equal_crs(self.geometry.crs, u_coord.NE_CRS) and self.geometry.size:
            return self.geometry
        self.set_geometry_points(scheduler)
        return self.geometry.to_crs(u_coord.NE_CRS)

    def __deepcopy__(self, memo):
        """Avoid error deep copy in gpd.GeoSeries by setting only the crs."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for key, value in self.__dict__.items():
            if key == 'geometry':
                setattr(result, key, gpd.GeoSeries(crs=self.geometry.crs))
            else:
                setattr(result, key, copy.deepcopy(value, memo))
        return result


def generate_nat_earth_centroids(res_as=360, path=None, dist_coast=False):
    """Generate hdf5 file containing Centroids of given resolution.

    For reproducibility, this is the function that generates the centroids files in
    `NATEARTH_CENTROIDS`. These files are provided with CLIMADA so that this function should never
    be called!

    Parameters
    ----------
    res_as : int
        Resolution of file in arc-seconds. Default: 360.
    path : str, optional
        If set, write resulting hdf5 file here instead of the default location.
    dist_coast : bool, optional
        If True, read distance from a NASA dataset (see util.coordinates.dist_to_coast_nasa).
        Default: False.
    """
    if path is None and res_as not in [150, 360]:
        raise ValueError("Only 150 and 360 arc-seconds are supported!")

    res_deg = res_as / 3600
    lat_dim = np.arange(-90 + res_deg, 90, res_deg)
    lon_dim = np.arange(-180 + res_deg, 180 + res_deg, res_deg)
    lon, lat = [ar.ravel() for ar in np.meshgrid(lon_dim, lat_dim)]
    natids = np.uint16(u_coord.get_country_code(lat, lon, gridded=False))

    cen = Centroids.from_lat_lon(lat, lon)
    cen.region_id = natids
    cen.set_lat_lon_to_meta()
    cen.lat = np.array([])
    cen.lon = np.array([])

    if path is None:
        path = NATEARTH_CENTROIDS[res_as]

    if dist_coast:
        cen.set_dist_coast(precomputed=True, signed=False)
        cen.dist_coast = np.float16(cen.dist_coast)
    cen.write_hdf5(path)
